char *macglyphorder[258] = {
  "notdef", ".null", "CR", "space", "exclam",
  "quotedbl", "numbersign", "dollar", "percent", "ampersand",
  "quotesingle", "parenleft", "parenright", "asterisk", "plus",
  "comma", "hyphen", "period", "slash", "zero",
  "one", "two", "three", "four", "five",
  "six", "seven", "eight", "nine", "colon",
  "semicolon", "less", "equal", "greater", "question",
  "at", "A", "B", "C", "D",
  "E", "F", "G", "H", "I",
  "J", "K", "L", "M", "N",
  "O", "P", "Q", "R", "S",
  "T", "U", "V", "W", "X",
  "Y", "Z", "bracketleft", "backslash", "bracketright",
  "asciicircum", "underscore", "grave", "a", "b",
  "c", "d", "e", "f", "g",
  "h", "i", "j", "k", "l",
  "m", "n", "o", "p", "q",
  "r", "s", "t", "u", "v",
  "w", "x", "y", "z", "braceleft",
  "bar", "braceright", "asciitilde", "Adieresis", "Aring",	/* 95-99 */
  "Ccedilla", "Eacute", "Ntilde", "Odieresis", "Udieresis",	/* 100-104 */
  "aacute", "agrave", "acircumflex", "adieresis", "atilde",	/* 105-109 */
  "aring", "ccedilla", "eacute", "egrave", "ecircumflex",	/* 110-114 */
  "edieresis", "iacute", "igrave", "icircumflex", "idieresis",	/* 115-119 */
  "ntilde", "oacute", "ograve", "ocircumflex", "odieresis",	/* 120-124 */
  "otilde", "uacute", "ugrave", "ucircumflex", "udieresis",	/* 125-129 */
  "dagger", "degree", "cent", "sterling", "section",		/* 130-134 */
  "bullet", "paragraph", "germandbls", "registered", "copyright", /* 135-139 */ 
  "trademark", "acute", "dieresis", "notequal", "AE",		/* 140-144 */
  "Oslash", "infinity", "plusminus", "lessequal", "greaterequal", /* 145-149 */
  "yen", "mu1", "partialdiff", "summation", "product",		/* 150-154 */
  "pi", "integral", "ordfeminine", "ordmasculine", "Ohm",	/* 155-159 */
  "ae", "oslash", "questiondown", "exclamdown", "logicalnot",	/* 160-164 */
  "radical", "florin", "approxequal", "increment", "guillemotleft", /* 165-169 */ 
  "guillemotright", "ellipsis", "nbspace", "Agrave", "Atilde",	/* 170-174 */
  "Otilde", "OE", "oe", "endash", "emdash",			/* 175-179 */
  "quotedblleft", "quotedblright", "quoteleft", "quoteright",	/* 180-184 */
  "divide",
  "lozenge", "ydieresis", "Ydieresis", "fraction", "currency",	/* 185-189 */
  "guilsinglleft", "guilsinglright", "fi", "fl", "daggerdbl",	/* 190-194 */
  "periodcentered", "quotesinglbase", "quotedblbase", "perthousand",  /* 195-199 */
  "Acircumflex",
  "Ecircumflex", "Aacute", "Edieresis", "Egrave", "Iacute",	/* 200-204 */ 
  "Icircumflex", "Idieresis", "Igrave", "Oacute", "Ocircumflex",/* 205-209 */
  "applelogo", "Ograve", "Uacute", "Ucircumflex", "Ugrave",	/* 210-214 */
  "dotlessi", "circumflex", "tilde", "overscore", "breve",	/* 215-219 */
  "dotaccent", "ring", "cedilla", "hungarumlaut", "ogonek",	/* 220-224 */
  "caron", "Lslash", "Islash", "Scaron", "scaron",		/* 225-229 */
  "Zcaron", "zcaron", "brokenbar", "Eth", "eth",		/* 230-234 */
  "Yacute", "yacture", "Thorn", "thorn", "minus",		/* 235-239 */
  "multiply", "onesuperior", "twosuperior", "threesuperior",
  "onehalf",							/* 240-244 */
  "onequarter", "threequarters", "franc", "Gbreve", "gbreve",	/* 245-249 */
  "ldot", "Scedilla", "scedilla", "Cacute", "cacute", 		/* 250-254 */
  "Ccaron", "ccaron", "dmacron"					/* 255-257 */
};
